#include "memchk.hpp"
#include "yyypostman.hpp"
#include "yyynetwork.hpp"
#include "fatal.hpp"
#include "yyyscheduler.hpp"
#include "yyyexceptions.hpp"
#include "yyysendables.hpp"
#include "display.hpp"
#include "messages.hpp"

#include <ncurses.h>

using namespace yyyscheduler;
using namespace yyypostman;
using namespace yyynetwork;
using namespace yyysmartptr;
using namespace yyybinbuff;

class clientprocess : public SigC::Object
{
public:
  typedef connection<scheduler,binbuff> con_t;
  typedef smartptr<con_t> sptr_con_t;
  typedef client<scheduler, binbuff> client_t;
  typedef postman<const uint8,con_t> postman_t;

  tempm_cell area;
  sptr_con_t mycon;
  client_t myclient;
  postman_t *post;
  scheduler::ionotify io;
  string input;

  tempm_cell::index_t cx, cy, enterline, ltx, lty;
  uint8 mycolor;

  clientprocess (int argc, char *argv[])
    : post(0)
  {
    if (argc!=2) fatal("must pass one parameter. ip address of server.");
    myclient.netconnect(ipv4address(argv[1]), portnumber(10000))
      .connect(SigC::slot(*this,&clientprocess::resultavailable));

    io.fire.connect(SigC::slot(*this,&clientprocess::keypressed));
    io.setfilehandle(0,scheduler::ionotify::READ);

    cx=cy=enterline=ltx=lty=0;
    mycolor=3;
    redraw();
  };

  ~clientprocess()
  {
    delete post; // <-- important that postman gets killed BEFORE
                 // the connection object
  };

  void resultavailable (client_t &, client_t::result &r)
  {
    if (r.isconnected()) {
      mycon=r.getconnection();
      mycon->sig_disconnected.connect
	(SigC::slot(*this,&clientprocess::disconnected));
      post=new postman_t(&(*mycon));
      post->corruptreceived.connect
	(SigC::slot(*this,&clientprocess::corruptreceived));
      post->sender1_open<const MSG_cc &>('x');
      post->receiver1_open<const MSG_cc &>('s')
	.connect(SigC::slot(*this,&clientprocess::recv_text));
      post->receiver1_open<const tempm_cell &>('a')
	.connect(SigC::slot(*this,&clientprocess::recv_area));
    }else{
      disp_deinit();
      printf("connection refused...\n");
      scheduler::gsched->setterminated(true);
    };
  };

  void disconnected(con_t &)
  {
    disp_deinit();
    printf("connection disconnected\n");
    scheduler::gsched->setterminated(true);
  };

  void corruptreceived(postman_t &)
  {
    disp_deinit();
    printf("postman corruptreceived\n");
    mycon->disconnect();
  };

  void recv_text(const MSG_cc &m)
  {
    if ((m.x<0) || (m.y<0) || (m.x>=area.xsize()) || (m.y>=area.ysize()))
      fatal("received out of area update request");
    area(m.x, m.y)=m.cl;
    redraw();
  };

  void recv_area(const tempm_cell &area_)
  {
    cx=cy=enterline=ltx=lty=0;
    area=area_;
    redraw();
  };

  void keypressed(scheduler::ionotify &)
  {
    int c=getch();
    switch (c) {
    case KEY_DOWN : cy++; break;
    case KEY_RIGHT : cx++; break;
    case KEY_UP: cy--; break;
    case KEY_LEFT: cx--; break;
    case '\r':
      cx=enterline;
      cy++;
      break;
    case KEY_BACKSPACE:
      if (cx>0) {
	cx--;
	cell myc;
	myc.ascii=' ';
	myc.colnum=1;
	modify(cx,cy,myc);
      };
      break;
    case KEY_F(1):
      enterline=cx;
      break;
    case KEY_F(2):
      if (mycolor > 1) mycolor--;
      break;
    case KEY_F(3):
      if (mycolor < 22) mycolor++;
      break;
    case KEY_F(10):
      scheduler::gsched->setterminated(true);
      break;
    default:
      if ((c>=32)&&(c<=127)) {
	cell myc;
	myc.ascii=c;
	myc.colnum=mycolor;
	modify(cx,cy,myc);
	cx++;
      };
    };

    cursorrepair();
    redraw();
  };

  void cursorrepair()
  {
    if (cy>=area.ysize()) cy=area.ysize()-1;
    if (cx>=area.xsize()) cx=area.xsize()-1;
    if (cy<0) cy=0;
    if (cx<0) cx=0;
  };

  void modify(const tempm_cell::index_t &px,
	      const tempm_cell::index_t &py,
	      const cell &c)
  {
    if ((px<0) || (py<0) || (px>=area.xsize()) || (py>=area.ysize())) return;
    area(px,py)=c;
    MSG_cc m;
    m.x=px;
    m.y=py;
    m.cl=c;
    post->sender1<const MSG_cc &> ('x') (m);
  };

  void redraw()
  {
    if (ltx>cx) ltx=cx;
    if (lty>cy) lty=cy;
    if (ltx<cx-COLS+1) ltx=cx-COLS+1;
    if (lty<cy-LINES+1) lty=cy-LINES+1;

    for (tempm_cell::index_t y=0;y<LINES;++y)
      for (tempm_cell::index_t x=0;x<COLS;++x) {
	tempm_cell::index_t rx=x+ltx, ry=y+lty;
	if ((rx<0) || (ry<0) || (rx>=area.xsize()) || (ry>=area.ysize())) {
	  disp_put(x,y,'?',3);
	}else{
	  disp_put(x,y,area(rx,ry).ascii,area(rx,ry).colnum);
	};
      };
    if (mycon.isnull())
      disp_putstring(0,0,"Connecting to server and getting data...",16);

    disp_curs(cx-ltx,cy-lty);
    disp_update();
  };
};

int main(int argc, char *argv[])
{
  memchk::resetsituation();
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      disp_init();
      clientprocess cli(argc,argv);
      s.run();
      disp_deinit();
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
