#include "memchk.hpp"
#include "yyypostman.hpp"
#include "yyynetwork.hpp"
#include "fatal.hpp"
#include "yyyscheduler.hpp"
#include "yyyexceptions.hpp"
#include "yyybinbuff.hpp"
#include "yyysendables.hpp"
#include "yyysigdeferer.hpp"
#include "messages.hpp"

#include <unistd.h>

using namespace yyybinbuff;
using namespace yyyscheduler;
using namespace yyypostman;
using namespace yyynetwork;
using namespace yyysmartptr;
using namespace yyysigdeferer;

class serverprocess;
class peer_t;

typedef connection<scheduler,binbuff> con_t;
typedef smartptr<con_t> sptr_con_t;
typedef server<scheduler,binbuff> server_t;
typedef postman<const uint8,con_t> postman_t;
typedef smartptr<postman_t> sptr_postman_t;
typedef smartptr<peer_t> sptr_peer_t;

class peer_t : public SigC::Object {
public:
  sptr_con_t con;
  sptr_postman_t post;
  serverprocess &sp;

  SigC::Signal1<void,peer_t*> killme;

  peer_t(serverprocess &sp_, sptr_con_t &c);

  ~peer_t()
  {
    cout << "destructor of peer ";
    debug();
    cout << " called.\n";
    post.kill(); // must be killed before it's connection!
  };

  void disconnected(con_t &)
  {
    debug();
    cout << " disconnected !\n";
    killme.emit(this);
  };

  void corruptreceived(postman_t &)
  {
    debug();
    cout << " postman corruptreceived, cutting connection ...\n";
    killme.emit(this);
  };

  void debug()
  {
    cout << '[' << con->getforeign_ipv4address().dump() << ']';
  };
};

class serverprocess : public SigC::Object
{
public:
  tempm_cell area;
  server_t myserver;
  scheduler::ionotify io;

  list<sptr_peer_t> mycons;
  sigdeferer1<peer_t*, scheduler> killpeerdefer;

  serverprocess (int argc, char *argv[])
  {
    if (argc!=3) fatal("must pass two parameters. x and y size of text area.");
    tempm_cell::index_t xs(atol(argv[1])), ys(atol(argv[2]));
    if ((xs<10) || (xs>1000) || (ys<10) || (ys>1000)) fatal("map size out of range. try 200 100 for a start.");
    area.resize(xs,ys);

    myserver.resultavailable.connect
      (SigC::slot(*this,&serverprocess::newconnection));
    myserver.setport(10000);

    io.fire.connect(SigC::slot(*this,&serverprocess::keypressed));
    io.setfilehandle(0,scheduler::ionotify::READ);

    killpeerdefer.send.connect(SigC::slot(*this,&serverprocess::killpeer));

    cout << "Enter 'q' to quit server\n";
  };

  void newconnection(server_t &, sptr_con_t &c)
  {
    sptr_peer_t np(new peer_t(*this,c));
    np->killme.connect(SigC::slot
		       (killpeerdefer,
			&sigdeferer1<peer_t*,scheduler>::receive));
    mycons.push_back(np);
  };

  void killpeer(peer_t *p)
  {
    list<sptr_peer_t>::iterator i;
    for (i=mycons.begin();i!=mycons.end();++i) {
      if (&(**i) == p) {
	mycons.erase(i);
	return;
      };
    };
    fatal("serverprocess::killpeer didn't find it's peer");
  };

  void keypressed(scheduler::ionotify &)
  {
    char c;
    read(0,&c,1);
    if (c=='q') {
      cout << "shutting down...\n";
      scheduler::gsched->setterminated(true);
    };
  };

  void recv_text(const MSG_cc &m)
  {
    if ( ! ((m.cl.ascii>=32)&&(m.cl.ascii<=127)) ) {
      cout << "recved illegal ascii code\n";
      return;
    };
    cout << "[" << int(m.x) << ',' << int(m.y) << ',' << m.cl.ascii << ',' << int(m.cl.colnum) << "]\n";
    if ((m.x<0) || (m.y<0) || (m.x>=area.xsize()) || (m.y>=area.ysize())) {
      cout << "that was illegal coordinates, ignoring !\n";
      return;
    };
    if ( ! ((m.cl.colnum>=1)&&(m.cl.colnum<=22)) ) {
      cout << "that was illegal color, ignoring !\n";
      return;
    };
    area(m.x,m.y)=m.cl;
    list<sptr_peer_t>::iterator i;
    for (i=mycons.begin();i!=mycons.end();++i)
      (*i)->post->sender1<const MSG_cc &> ('s') (m);
  };
};





peer_t::peer_t(serverprocess &sp_, sptr_con_t &c)
  : sp(sp_)
{
  con=c;
  
  con->sig_disconnected.connect
    (SigC::slot(*this,&peer_t::disconnected));
  
  post=sptr_postman_t(new postman_t(&(*con)));
  
  post->corruptreceived.connect
    (SigC::slot(*this,&peer_t::corruptreceived));
  
  post->receiver1_open<const MSG_cc &> ('x')
    .connect(SigC::slot(sp,&serverprocess::recv_text));

  post->sender1_open<const MSG_cc &> ('s');

  post->sender1_open<const tempm_cell &> ('a');
  post->sender1<const tempm_cell &> ('a') (sp.area);
  
  cout << "new connection from ";
  debug();
  cout << ".\n";
};





int main(int argc, char *argv[])
{
  memchk::resetsituation();
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      serverprocess serv(argc,argv);
      s.run();
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
