/*

  yyylowlevelnet - a low level networking library
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "yyylowlevelnet.hpp"

#include "fatal.hpp"

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

using namespace yyynetwork;

// ------ hostname
hostname::hostname()
  : isdefined_rep(false)
{
};

hostname::hostname (const string &s)
  : rep(s), isdefined_rep(true)
{
};

const string hostname::getname() const
{
  return rep;
};

const bool hostname::isdefined() const
{
  return isdefined_rep;
};

void hostname::undefine ()
{
  rep.erase();
  isdefined_rep = false;
};


// ------ ipv4address
ipv4address::ipv4address ()
  : isdefined_rep(false)
{
};

ipv4address::ipv4address (const string &s)
{
  memset(&rep,0,sizeof rep);
  isdefined_rep = (inet_aton(s.c_str(),&rep)==0) ? false : true;
};

const string ipv4address::dump () const
{
  if (!isdefined()) fatal("ipv4address::dump: not defined address");
  return string(inet_ntoa(rep));
};

const struct in_addr & ipv4address::getraw () const
{
  if (!isdefined_rep) fatal("ipv4address::getraw: not defined address");
  return rep;
};

void ipv4address::setraw (const struct in_addr & i)
{
  memcpy(&rep,&i,sizeof rep);
  isdefined_rep=true;
};

const bool ipv4address::isdefined () const
{
  return isdefined_rep;
};

void ipv4address::undefine ()
{
  isdefined_rep = false;
};


// ------ portnumber
portnumber::portnumber ()
  : isdefined_rep(false)
{
};

portnumber::portnumber (const uint16 pn)
  : rep(pn), isdefined_rep(true)
{
};

const uint16 portnumber::getnumber () const
{
  if (!isdefined_rep) fatal("portnumber::getnumber: not defined port");
  return rep;
};

void portnumber::setnumber (const uint16 pn)
{
  rep=pn;
};

const bool portnumber::isdefined () const
{
  return isdefined_rep;
};

void portnumber::undefine ()
{
  isdefined_rep = false;
};

// ------ tcphandle
tcphandle::tcphandle ()
  : isopen_rep(false)
{
};

tcphandle::~tcphandle()
{
  close();
};

const bool tcphandle::isopen () const
{
  return isopen_rep;
};

void tcphandle::close ()
{
  if (isopen()) {
    if (mustshutdown) shutdown (handle_rep,2);
    // don't check for error, since connection can be a not connected socket which happens if write failed cause other end closed.
    if (::close (handle_rep)<0) fatal_errno("yyynetwork::tcphandle::~tcphandle: close() failed");
  };
  isopen_rep=false;
};

const int tcphandle::getfilehandle () const
{
  if (!isopen()) fatal("tcphandle::getfilehandle: not open");
  return handle_rep;
};

void tcphandle::makesocket()
{
  if (isopen()) fatal("tcphandle::makesocket: tcphandle already open");
  if ((handle_rep=socket(PF_INET,SOCK_STREAM,0))<0) fatal_errno ("tcphandle::makesocket: socket() failed");
  isopen_rep=true;
  mustshutdown=false;
};

void tcphandle::setreuse()
{
  if (!isopen()) fatal("tcphandle::setreuse: tcphandle not open");
  int val=1;
  if (setsockopt(handle_rep,SOL_SOCKET,SO_REUSEADDR,&val,sizeof val)<0)
    fatal_errno("tcphandle::setreuse: setsockopt failed");
};

void tcphandle::setsolinger()
{
  if (!isopen()) fatal("tcphandle::setsolinger: tcphandle not open");
  struct linger val;
  val.l_onoff=0;
  val.l_linger=0;
  if (setsockopt(handle_rep,SOL_SOCKET,SO_LINGER,&val,sizeof val)<0)
    fatal_errno("tcphandle::setsolinger(): setsockopt failed");
};

void tcphandle::setnonblocking ()
{
  if (!isopen()) fatal("tcphandle::setnonblocking: tcphandle not open");
  if (fcntl(handle_rep,F_SETFL,O_NONBLOCK)<0) fatal_errno("tcphandle::setnonblocking: fcntl(...) failed");
};

void tcphandle::connect(const ipv4address &foreign_ipv4address, const portnumber &foreign_portnumber,
			connect_result_t &res)
{
  if (!isopen()) fatal("tcphandle::connect: tcphandle not open");

  struct sockaddr_in name;
  memset(&name,0,sizeof name);
  name.sin_family=AF_INET;
  memcpy(&name.sin_addr.s_addr, &foreign_ipv4address.getraw(), sizeof name.sin_addr.s_addr);
  name.sin_port=htons(foreign_portnumber.getnumber());

  int retval=::connect (handle_rep,(struct sockaddr *) &name,sizeof name);
  if (retval<0) {
    if (errno==EINPROGRESS) {
      res=INPROGRESS;
    }else if (errno==ECONNREFUSED) {
      res=REFUSED;
    }else if ((errno=ETIMEDOUT)||(errno=ENETUNREACH)||(errno==EHOSTUNREACH)) {
      res=UNREACHABLE;
    }else fatal_errno("tcphandle::connect: ::connect failed");
  }else{
    res=CONNECTED;
    mustshutdown=true;
  };
};

void tcphandle::bind(const portnumber &local_portnumber, bind_result_t &res)
{
  struct sockaddr_in name;
  memset(&name,0,sizeof name);
  name.sin_family = AF_INET;
  name.sin_port = htons (local_portnumber.getnumber());
  name.sin_addr.s_addr = htonl (INADDR_ANY);
  if (::bind (handle_rep, (struct sockaddr *) &name, sizeof (name)) < 0) {
    if (errno==EADDRINUSE) res = ADDRESSINUSE;
    else if (errno==EACCES) res = ACCESSDENIED;
    else fatal_errno("tcphandle::bind: ::bind failed");
  }else res=BINDED;
};

void tcphandle::listen ()
{
  if (!isopen()) fatal("tcphandle::listen: tcphandle not open");
  if (::listen(handle_rep,10)<0) fatal_errno("tcphandle::listen: ::listen failed");
};

void tcphandle::accept(tcphandle &newhandle, ipv4address &foreign_ipv4address, portnumber &foreign_portnumber,
		       accept_result_t &res)
{
  if (!isopen()) fatal("tcphandle::accept: tcphandle not open");
  if (newhandle.isopen()) fatal("tcphandle::accept: newhandle mustn't be open");
  struct sockaddr_in name;
  memset(&name,0,sizeof name);
  size_t s = sizeof name;
  int retcode=::accept(handle_rep,(struct sockaddr *) &name,&s);
  if (retcode<0) {
    if (errno==EWOULDBLOCK) res=WOULDBLOCK;
    else fatal_errno("tcphandle::accept: ::accept failed");
  }else{
    newhandle.handle_rep=retcode;
    newhandle.isopen_rep=true;
    newhandle.mustshutdown=true;
    foreign_ipv4address.setraw(name.sin_addr);
    foreign_portnumber.setnumber(htons(name.sin_port));
    res=ACCEPTED;
  };
};

const ssize_t tcphandle::read(void *const buf, const size_t count)
{
  if (!isopen()) fatal("tcphandle::read: tcphandle not open");
  return ::read(handle_rep,buf,count);
};

const ssize_t tcphandle::write(void *const buf, const size_t count)
{
  if (!isopen()) fatal("tcphandle::write: tcphandle not open");
  return ::write(handle_rep,buf,count);
};
